"""Annotations."""
import logging
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

from roux.stat.transform import rescale

def _set_text(
    ax: plt.Axes,
    ha: str='center',
    va: str='center',
    # off_text=0,
    **kws,
    ) -> plt.Axes:
    """Set text on the plot

    Args:
        ax (plt.Axes): subplot
        ha (str, optional): horizontal alignment. Defaults to 'center'.
        va (str, optional): vertical alignment. Defaults to 'center'.

    Returns:
        plt.Axes: subplot
    """
    if ax.name!='polar':
        ax.text(
            # y=0,
            rotation=90,
            ha=ha,
            va=va,
            **kws,
        )    
    else:
        angle=kws['x']
        rotation = np.rad2deg(angle)

        if angle <= np.pi/2:
            # alignment = "center"
            rotation = rotation + 180 + 2*abs(rotation-90)            
        elif angle < 3*(np.pi/2):
            # alignment = "center"
            rotation = rotation - 2*abs(rotation-90)
        else: 
            # alignment = "center"
            rotation = rotation + 180 -  2*abs(rotation-270)#2*abs(rotation-90)# + 2*abs(rotation-90)            
            
        # logging.info(f"{angle:.1f} {ha},{va},{rotation}")
        ax.text(
            ha=ha, 
            va=va, 
            rotation=rotation, 
            rotation_mode="anchor",
            **kws,
        )
    return ax

def annot_labels(
    ax_chrom: plt.Axes, # A
    data: pd.DataFrame, # coordinates
    colx: str, # coordinates data
    chrom_y: float,
    col_label: str,
    loc: str='out', # out in
    col_start: str=None, # coordinates chrosomome plots
    ax: plt.Axes=None, # B:data
    coly: str=None, # data
    col_labelx: str='label x', 
    color: str='darkgray',
    yoff_scales: float=None,
    off_labels_segments: float=20,
    scale_polar: float=1.5,
    fig: plt.Figure=None,
    test: bool=False,
    ) -> plt.Axes:
    """Annot labels e.g. gene names

    Args:
        ax_chrom (plt.Axes): subplot with the chromosome plot
        col_label (str): column with the labels
        loc (str, optional): locations. Defaults to 'out'.
        color (str, optional): color. Defaults to 'darkgray'.
        yoff_scales (float, optional): y offset. Defaults to None.
        off_labels_segments (float, optional): offset for the label segments. Defaults to 20.
        scale_polar (float, optional): scale for the polar plot. Defaults to 1.5.
        fig (plt.Figure, optional): figure. Defaults to None.
        test (bool, optional): test-mode. Defaults to False.

    Returns:
        plt.Axes: subplot
    """
    if fig is None:
        fig=plt.gcf()                
    if col_start is None:
        col_start=colx
    ## labels
    ### rescale coordinates
    # data=data.assign(**{colx:lambda df: to_polar(df[colx],range1=[df[colx].min(),df[colx].max()],range2=xlim)})
    if ax_chrom.name != 'polar':
        df1=(data
            .sort_values(colx)
            .assign(**{col_labelx:lambda df: rescale(df[colx].rank(),range2=ax_chrom.get_xlim())})    
            )
    else:
        df1=(data
            .sort_values(colx)
            .assign(**{
                f'{colx} quadrant': lambda df : pd.cut(df[colx],bins=np.linspace(ax_chrom.get_xlim()[0],ax_chrom.get_xlim()[1],
                                                                                 off_labels_segments, ## shrink the spacing between the labels
                                                                                ),include_lowest=True),
            })
        )
        l1=[]
        for name,df in df1.groupby(f'{colx} quadrant',observed=False):
            # print(name.left,name.right,df[colx].nunique())
            if df[colx].nunique()==0:
                continue
            elif df[colx].nunique()==1:
                l1+=df[colx].tolist()
                continue
            range2=(name.left,name.right-((name.right-name.left)/df[colx].nunique()))
            # print(range2)
            l1+=rescale(
                    df[colx].rank(),
                    range2=range2
                       ).tolist()
            del range2
        assert len(l1)==len(df1)
        df1[col_labelx]=l1
        del l1
        # print(df1)
    ## annotations
    # scaled by the distance between the data plot and the cromosomes
    dist=chrom_y-ax_chrom.get_ylim()[0]
    chrom_yoff=-0.1 if loc=='out' else 0
    line1B_yoff=dist*(0.03 if loc=='out' else 0.95) if yoff_scales is None else yoff_scales[0]
    line2B_yoff=dist*(0.06 if loc=='out' else 0.85) if yoff_scales is None else yoff_scales[1]
    ## because the polar plots overlap, rectilinear stack
    line3A_yoff=dist*(0.09 if loc=='out' else 0.65) if yoff_scales is None else yoff_scales[2]
    # if not coly is None:
    #     line3B_yoff= data[coly].max()## TODO
    line3B_yoff=0
    if loc=='out':
        label_yoff =dist*(0.12*(scale_polar if ax_chrom.name=='polar' else 0.8) if yoff_scales is None else yoff_scales[3])
    else:
        label_yoff =(line3A_yoff+line2B_yoff)/2
    # print(dist*0.12)
    # print(f"label_yoff={label_yoff}")
    
    from matplotlib.patches import ConnectionPatch
    if loc=='out':
        # default
        # # -..to gene
        if test:
            logging.info(df1)
        _=df1.apply(lambda x:  fig.add_artist(
            ConnectionPatch(
                xyA=[x[col_start],chrom_y+chrom_yoff], 
                xyB=[x[col_start],ax_chrom.get_ylim()[1]+line1B_yoff], 
                coordsA=ax_chrom.transData,#"data", 
                coordsB=ax_chrom.transData,#"data",
                # axesA=ax_chrom,
                # axesB=ax,
                color=color,
                alpha=0.5,
                zorder=2,
                clip_on=False
            ),
            ),
            axis=1)
        # .-. chromosome
        _=df1.apply(lambda x:  fig.add_artist(
            ConnectionPatch(
                xyA=[x[col_start],ax_chrom.get_ylim()[1]+line1B_yoff], 
                xyB=[x[col_labelx],ax_chrom.get_ylim()[1]+line2B_yoff], 
                coordsA=ax_chrom.transData,#"data", 
                coordsB=ax_chrom.transData,#"data",
                # axesA=ax_chrom,
                # axesB=ax,
                color=color,
                alpha=0.5,
                zorder=2,
                clip_on=False
            ),
            ),
            axis=1)
        _=df1.apply(lambda x: fig.add_artist(
            ConnectionPatch(
                xyA=[x[col_labelx],ax_chrom.get_ylim()[1]+line2B_yoff], 
                xyB=[x[col_labelx],ax_chrom.get_ylim()[1]+line3A_yoff], 
                coordsA=ax_chrom.transData,#"data", 
                coordsB=ax_chrom.transData,#"data",
                # axesA=ax_chrom,
                # axesB=ax,
                color=color,
                alpha=0.5,
                zorder=2,
                clip_on=False
            ),
            ),
            axis=1)
        #..- to data
        if not coly is None:
            _=df1.apply(lambda x:  fig.add_artist(
                ConnectionPatch(
                    xyA=[x[col_start],chrom_y+chrom_yoff], 
                    xyB=[x[colx],ax.get_ylim()[1]+line3B_yoff], 
                    coordsA=ax_chrom.transData,#"data", 
                    coordsB=ax.transData,#"data",
                    axesA=ax_chrom,
                    axesB=ax,
                    color=color,
                    alpha=0.5,
                    zorder=2,
                    clip_on=False
                ),
                ),
                axis=1)
            _=df1.apply(lambda x:  fig.add_artist(
                ConnectionPatch(
                    xyA=[x[colx],ax.get_ylim()[1]+line3B_yoff], 
                    xyB=[x[colx],x[coly]], 
                    coordsA=ax.transData,#"data", 
                    coordsB=ax.transData,#"data",
                    axesA=ax,
                    axesB=ax,
                    color=color,
                    alpha=0.5,
                    zorder=2,
                    clip_on=False
                ),
                ),
                axis=1)
    else:
        # # -.. chromosome
        if test:
            logging.info(df1)
        _=df1.apply(lambda x:  fig.add_artist(
            ConnectionPatch(
                xyA=[x[col_start],chrom_y+chrom_yoff], 
                xyB=[x[col_start],ax_chrom.get_ylim()[0]+line1B_yoff], 
                coordsA=ax_chrom.transData,#"data", 
                coordsB=ax_chrom.transData,#"data",
                # axesA=ax_chrom,
                # axesB=ax,
                color=color,
                alpha=0.5,
                zorder=2,
                clip_on=False
            ),
            ),
            axis=1)
        # .-. gene
        _=df1.apply(lambda x:  fig.add_artist(
            ConnectionPatch(
                xyA=[x[col_start],ax_chrom.get_ylim()[0]+line1B_yoff], 
                xyB=[x[col_labelx],ax_chrom.get_ylim()[0]+line2B_yoff], 
                coordsA=ax_chrom.transData,#"data", 
                coordsB=ax_chrom.transData,#"data",
                # axesA=ax_chrom,
                # axesB=ax,
                color=color,
                alpha=0.5,
                zorder=2,
                clip_on=False
            ),
            ),
            axis=1)
        _=df1.apply(lambda x:  fig.add_artist(
            ConnectionPatch(
                xyA=[x[col_labelx],ax_chrom.get_ylim()[0]+line2B_yoff], 
                xyB=[x[col_labelx],ax_chrom.get_ylim()[0]+line3A_yoff], 
                coordsA=ax_chrom.transData,#"data", 
                coordsB=ax_chrom.transData,#"data",
                # axesA=ax_chrom,
                # axesB=ax,
                color=color,
                alpha=0.5,
                zorder=2,
                clip_on=False
            ),
            ),
            axis=1)
        # ..- data
        if not coly is None:
            _=df1.apply(lambda x:  fig.add_artist(
                ConnectionPatch(
                    xyA=[x[col_labelx],ax_chrom.get_ylim()[0]+line3A_yoff], 
                    xyB=[x[colx],ax.get_ylim()[1]+line3B_yoff], 
                    coordsA=ax_chrom.transData,#"data", 
                    coordsB=ax.transData,#"data",
                    axesA=ax_chrom,
                    axesB=ax,
                    color=color,
                    alpha=0.5,
                    zorder=2,
                    clip_on=False
                ),
                ),
                axis=1)
            _=df1.apply(lambda x:  fig.add_artist(
                ConnectionPatch(
                    xyA=[x[colx],ax.get_ylim()[1]+line3B_yoff], 
                    xyB=[x[colx],x[coly]], 
                    coordsA=ax.transData,#"data", 
                    coordsB=ax.transData,#"data",
                    axesA=ax,
                    axesB=ax,
                    color=color,
                    alpha=0.5,
                    zorder=2,
                    clip_on=False
                ),
                ),
                axis=1)
    if ax is None:
        ax=ax_chrom
    if ax.name !='polar':
        ax_chrom.axhspan(
            ymin=ax_chrom.get_ylim()[0],
            ymax=0,
             color='w',
             # zorder=0,
            )
    else:
        # print(
        #     ax_chrom.get_ylim(),
        #     ax_chrom.get_ylim()[1]-ax_chrom.get_ylim()[0],
        #     ax_chrom.get_ylim()[0]+line1B_yoff,
        #     (ax_chrom.get_ylim()[0]+line1B_yoff)-ax_chrom.get_ylim()[0],
        #     ((ax_chrom.get_ylim()[0]+line1B_yoff)-ax_chrom.get_ylim()[0])/(ax_chrom.get_ylim()[1]-ax_chrom.get_ylim()[0]),
        # )
        from chrov.viz.chrom import _set_span_polar 
        # print(ax_chrom.get_ylim())
        # print(((ax_chrom.get_ylim()[0]+line1B_yoff)-ax_chrom.get_ylim()[0])/(ax_chrom.get_ylim()[1]-ax_chrom.get_ylim()[0]))
        _set_span_polar(
            ax=ax_chrom,
            span_type='v',
            start=ax.get_xlim()[0],
            end=ax.get_xlim()[1],
            intervals=100,
            # ymin=0 if loc!='out' else ((ax_chrom.get_ylim()[0]+line1B_yoff)-ax_chrom.get_ylim()[0])/(ax_chrom.get_ylim()[1]-ax_chrom.get_ylim()[0]),
            ymax=0.85,# if loc=='out' else ((ax_chrom.get_ylim()[0])-ax_chrom.get_ylim()[0])/(ax_chrom.get_ylim()[1]-ax_chrom.get_ylim()[0]),
            color='w',
            )
    _=df1.drop_duplicates(subset=[col_label,col_labelx]).apply(lambda x: _set_text(
        ax=ax_chrom,
        x=x[col_labelx],
        y=(ax_chrom.get_ylim()[1 if loc=='out' else 0]+label_yoff),
        s=x[col_label],
        ha='center',# if loc!='out' and not ax_chrom.name=='polar' else 'left',
        va='bottom' if loc=='out' and ax_chrom.name!='polar' else 'center',
        ),axis=1)
    if ax.name == 'polar':
        ax.set(
            # xlabel=None,
            ylabel=None
        )
    return ax
    
def show_segments(
    ax: plt.Axes,
    y: float,
    offy: float,
    kind: str='arrows',
    segments: dict=None,
    segments_kws: dict={},
    offytext: float=0.2,
    arrow_kws: dict=dict(
        color='k',
        lw=1,    
        alpha=1,
        arrowstyle='<->',
        ),
    test: bool=False,
    **kws_annotate,
    )->plt.Axes:
    """Show segments aligned to chromosome arm.

    Args:
        ax (plt.Axes): subplot
        data (pd.DataFrame): input data
        size (int): size of the segments
    """    
    if segments is None:
        ## get the boundaries
        from roux.lib.set import get_windows
        df0=get_windows(
            **segments_kws,
            )
    else:
        def to_df(segments):
            if isinstance(segments,list):
                df0=pd.DataFrame(segments,columns=['start','end']).sort_values('start').reset_index(drop=True)
                df0.index.name='label'
                df0=df0.reset_index().assign(label=lambda df: df['label'].apply(lambda x: f"segment{x+1}"))
            elif isinstance(segments,dict):            
                df0=pd.DataFrame(segments,
                    index=['start','end']
                    ).T
            elif isinstance(segments,pd.DataFrame):            
                df0=segments
            else:
                raise ValueError(segments)
            ## post-processing
            return df0 
        df0=to_df(segments)
    if kind=='arrows':
        _=(df0
            .apply(lambda x: ax.annotate(
                text='', 
                xy=(x['start'],
                    y if (x.name%2)==0 else y-offy), 
                xytext=(x['end'],
                        y if (x.name%2)==0 else y-offy), 
                arrowprops=dict(
                    shrinkA=0,shrinkB=0,
                    **arrow_kws,
            ),
            **kws_annotate,
            ),
            axis=1)
        )
    elif kind=='bands':
        _=df0.apply(lambda x: ax.plot(
                [x['start'],x['end']],
                [y if (x['window#']%2)==0 else y-offy,y if (x['window#']%2)==0 else y-offy],
                solid_capstyle='butt',
                lw=10,#color='gray',
                # alpha=0.3,
            ),
            axis=1)
    else:
        raise ValueError(kind) 
        
    ## labels    
    if 'label' in df0:
        _=(df0
            .apply(lambda x: ax.text(
                x=np.mean([x['start'],x['end']]),
                y=y+offytext,
                s=x['label'] if 'label' in x else '', 
                ha='center',
                va='bottom',
            ),
            axis=1)
        )
    if not test:
        ax.axis('off')
    return ax    



import numpy as np
import pandas as pd

import matplotlib.pyplot as plt

import roux.lib.dfs as rd

from chrov.viz.chrom import plot_arm

def plot_ranges(
    data:pd.DataFrame,
    col_id: str,
    col_start: str,
    col_end: str,
    start: int=None,
    end: int= None,
    hue: str= None,
    y=None, # rank
    kind=None,
    strand: str='+',
    cytobands: dict=None,
    cytobands_y: float = None,
    col_groupby: str=None,
    group_x: float=-0.22,
    # col_sortby: str=None,
    sort_ascending: bool=False,
    show_labels: str= True,
    col_label: str=None,
    col_label_right: str=None,
    color: str='b',
    colors: dict=None,
    palette: str= 'Set2',
    hue_lim: list= [],
    lw: int = 10,
    zorders: dict=None,
    show_segments: bool=False,
    xtick_interval:float=None,
    unit: str='Coordinate (bp)',
    test: bool=False,
    kws_legend:dict={},
    kws_hlines:dict={}, 
    ax: plt.Axes=None,
    )-> plt.Axes:
    """Plot ranges.

    Args:
        data (pd.DataFrame): input data.
        col_id (str): column with ids.
        col_start (str): column with start co-ordinates.
        col_end (str): column with end co-ordinates.
        end (int): end position for the plot
        start (int, optional): start position for the plot. Defaults to 0.
        hue (str, optional): column with color. Defaults to None.
        y (_type_, optional): column with y positions. Defaults to None.
        cytobands (dict, optional): cytobands to plot the chromosomes. Defaults to None.
        cytobands_y (float, optional): cytobands y-position. Defaults to None.
        col_groupby (str, optional): column to group by. Defaults to None.
        col_sortby (str, optional): column to sort by. Defaults to None.
        col_label (str, optional): column with labels. Defaults to None.
        col_label_right (str, optional): column with labels to be shown on the right side of the ranges. Defaults to None.
        colors (dict, optional): colors. Defaults to None.
        lw (int, optional): line width. Defaults to 10.
        zorders (dict, optional): z-orders. Defaults to None.
        show_segments (bool, optional): show segments. Defaults to False.
        xtick_interval (float, optional): x tick intervals. Defaults to None.
        test (bool, optional): test-mode. Defaults to False.
        ax (plt.Axes, optional): subplot. Defaults to None.

    Raises:
        ValueError: if kind is not 'split','separate' or 'joined'

    Returns:
        plt.Axes: subplot
    """
    assert len(data)!=0
    
    if ax is None:
        # fig,ax=plt.subplots(figsize=[3,len(data)*0.05])
        ax=plt.gca()
    from roux.viz.colors import saturate_color
    from roux.viz.ax_ import set_legend_custom
    
    if start is None:
        start=data[col_start].min()
    if end is None:
        end=data[col_end].max()

    assert not all(data[col_start]==data[col_end]), "lengths of all ranges is 0."
    ## trim data
    df1=(data
        .log.dropna(
            subset=[col_id,col_start,col_end]
        )
        # .sort_values(col_start,ascending=False)
        .assign(
        **{
            col_start: lambda df: df[col_start].apply(lambda x: x if x >= start else start), # trim
            col_end: lambda df: df[col_end].apply(lambda x: x if x <= end else end), # trim
        }
    ))    
    ## test for unique ids
    if kind is not None:
        if kind.lower().startswith('join'):
            col_label=None
    if col_label is None:
        col_label=col_id
        
    # if not col_sortby is None:
    #     df1=df1.sort_values(col_sortby,ascending=sort_ascending)
        
    if y is None:
        y='y'
        if kind in [None,'split','separate']: 
            df1=(df1
                .rd.assert_no_dups(col_id)
                .assign(
                    **{
                        y:np.arange(len(df1)),
                    }
                )
                )
        elif kind=='joined':
            df1=(
                df1
                .assign(
                    **{
                        y:lambda df: df.groupby(col_id,sort=False)[col_id].ngroup()
                    }#.transform(lambda x: range(len(x))))
                    )
            )
        else:
            raise ValueError(kind)
        assert y in df1
    if not hue is None:
        if df1[hue].dtype=='float':
            from roux.viz.colors import get_val2color
            # get_val2color?
            colors,colors_legend=get_val2color(
                df1[hue],
                vmin=hue_lim[0],
                vmax=hue_lim[1],
                cmap=palette,
            )
        else:
            if colors is None:
                from roux.viz.colors import get_ncolors
                colors=get_ncolors(
                    df1[hue].nunique(),
                    cmap=palette,
                    )
            if isinstance(colors,list):
                colors=dict(zip(df1[hue].unique(),colors))        
    if zorders is None:
        zorders={}

    ## lines
    _=df1.apply(lambda x: 
        ax.hlines(
            y=x[y],
            xmin=x[col_start],
            xmax=x[col_end],
            color=colors[x[hue]] if not hue is None else color,
            lw=lw,
            **kws_hlines,
            ) if not pd.isnull(x[y]) else None,
            axis=1,
           )
    ax.invert_yaxis()
    # labels
    if show_labels:
        if kind in [None,'split','separate']:            
            _=df1.apply(lambda x: ax.text(
                x=x[col_start],
                y=x[y],
                s=f"{x[col_label]} ",
                ha='right',va='center',
                # color=x['label_color'],
                ) if not pd.isnull(x[y]) else None,
            axis=1)
            if not col_label_right is None:
                _=df1.apply(lambda x: ax.text(
                        x=x[col_end],
                        y=x[y],
                        s=f"{x[col_label_right]} ",
                        ha='left',
                        va='center',
                        # color=x['label_color'],
                        ),
                        axis=1,
                       )                
        elif kind.lower().startswith('join'):
            _=(
                df1
                .loc[:,[col_label,y]]
                .drop_duplicates()
                .apply(lambda x: ax.text(
                    x=start if strand!='-' else end,
                    y=x[y],
                    s=f"{x[col_label]} ",
                    ha='right',
                    va='center',
                    # color=x['label_color'],
                    ),
                    axis=1)
              )        
    # _=df1.apply(lambda x: ax.plot(
    #     [x['start'],x['end']],[x[y],x[y]],
    #     color=colors[x[hue]] if x[hue] in colors else 'w',
    #     # alpha=1,
    #     solid_capstyle='butt',
    #     # zorder=2,
    #     zorder=zorders[x[hue]] if x[hue] in zorders else 2,
    #     lw=lw,
    #     ),
    #     axis=1)
    if not cytobands is None:
        if cytobands_y is None:
            cytobands_y=df1['y'].max()+1
        plot_arm(
            cytobands,
            y=cytobands_y,
            lw=lw,
            ax=ax,
            test=True,
        )
    if show_segments:
        from chrov.viz.chr import plot_segments
        plot_segments(ax)

    if not hue is None:    
        set_legend_custom(
            ax,
            legend2param=colors_legend if df1[hue].dtype=='float' else colors,
            param='color',
            lw=lw,
            marker=None,
            markerfacecolor=True,
            size=10,
            color='k',
            linestyle='-',
            **{
                **dict(
                    title=hue,
                    frameon=False,
                    bbox_to_anchor=[0.5,0],
                    loc='upper center',
                    ncol=3,
                    handlelength=0.1,
                ),
                **kws_legend,
            }
        )
    
    ## show groups
    if col_groupby in df1:
        df_=df1.loc[:,[y,col_groupby]].drop_duplicates()
        ax.set(
            yticks=df_[y].tolist(),
            yticklabels=df_.apply(
                lambda x: f"{x[col_groupby]}:{x[y]}",
                axis=1,
            ),
        )
        from roux.viz.ax_ import split_ticklabels
        split_ticklabels(
            ax=ax,
            axis='y',
            fmt='group',
            sep=":",
            group_x=group_x, ## todo: infer
            test=test,
        )
    # else:
    if not test:
        ax.axes.yaxis.set_visible(test)
    
    ax.xaxis.tick_top()
    ax.xaxis.set_label_position('top')
    ax.spines['top'].set_color('k')
    if not cytobands_y is None:
        ax.spines['top'].set_position(('data', cytobands_y+0.25))
        
    # removing the ax box lines
    _=plt.setp(
        ax.spines.values(),
        zorder=0,
        color='none',
    )
    # _=plt.setp([ax.get_xticklines(), ax.get_yticklines()], zorder=0)
    
    if not xtick_interval is None:
        ## set tick intervals
        import matplotlib.ticker as plticker
        loc = plticker.MultipleLocator(base=xtick_interval) # this locator puts ticks at regular intervals
        ax.xaxis.set_major_locator(loc)
    
    ax.set(
        xlim=[start,end],
        ylim=[df1[y].min()-0.5,df1[y].max()+1.5],
        xlabel=unit,
    )
    if strand=='-':
        ax.invert_xaxis()    
    return ax


from testbook import testbook
    
import os
from os import makedirs
from os.path import exists
assert len(set(['chrov','examples','tests']) - set(os.listdir('.')))==0
assert exists('./examples/inputs')
makedirs('./examples/outputs',exist_ok=True)
    
@testbook('examples/chrov_viz_annot.ipynb', execute=True)
def test_chrov_viz_annot(tb):
    pass # execute only because tests are present in the notebook itself
    return

@testbook('examples/chrov_viz_chrom.ipynb', execute=True)
def test_chrov_viz_chrom(tb):
    pass # execute only because tests are present in the notebook itself
    return

@testbook('examples/chrov_viz_chrom_custom.ipynb', execute=True)
def test_chrov_viz_chrom_custom(tb):
    pass # execute only because tests are present in the notebook itself
    return

@testbook('examples/chrov_viz_figure_custom.ipynb', execute=True)
def test_chrov_viz_figure_custom(tb):
    pass # execute only because tests are present in the notebook itself
    return

@testbook('examples/chrov_viz_figure_stem.ipynb', execute=True)
def test_chrov_viz_figure_stem(tb):
    pass # execute only because tests are present in the notebook itself
    return

@testbook('examples/chrov_viz_ranges.ipynb', execute=True)
def test_chrov_viz_ranges(tb):
    pass # execute only because tests are present in the notebook itself
    return

@testbook('examples/chrov_viz_tests.ipynb', execute=True)
def test_chrov_viz_tests(tb):
    pass # execute only because tests are present in the notebook itself
    return


import pandas as pd
import matplotlib.pyplot as plt

from chrov.viz.plot import plot_seaborn
from chrov.viz.chrom import annot_chroms, _concat_chroms
from chrov.viz.annot import annot_labels

def plot_with_chroms(
    data: pd.DataFrame,
    cytobands: pd.DataFrame,
    kind: str,
    colx: str,
    coly: str,
    col_label: str,
    va: str,
    col_start: str=None,
    xkind: str='loci', # colx contains coordinates
    coffy: str=None,
    off: float=None,
    offy: float=None,
    chrom_y: float=0,
    show_genome: bool=False,
    arc: bool=True,
    pi_span: float=1,
    pi_start: int=0,
    pi_end: int=None,
    fig: plt.Figure=None,
    figsize: list=None,
    ax_data: plt.Axes=None,
    kws_seaborn: dict={},
    kws_annot_chroms: dict={},
    kws_annot_labels: dict={},
    test: bool=False,
    ) -> plt.Figure:
    """Plot with chromosomes.

    Args:
        data (pd.DataFrame): input table
        cytobands (pd.DataFrame): cytobands
        kind (str): kind of plot
        colx (str): column with x values
        coly (str): column with y values
        col_label (str): column with labels
        va (str): vertical alignment
        col_start (str, optional): column with start positions. Defaults to None.
        xkind (str, optional): kind of x values. Defaults to 'loci'.
        off (float, optional): offset scale of the chromosome plot. Defaults to None.
        offy (float, optional): offset y of the chromosome plot. Defaults to None.
        chrom_y (float, optional): chromosome y-position. Defaults to 0.
        arc (bool, optional): arc/polar plot or linear/rectangular one. Defaults to True.
        pi_span (float, optional): pi span. Defaults to 1.
        pi_start (int, optional): pi start. Defaults to 0.
        pi_end (int, optional): pi end. Defaults to None.
        fig (plt.Figure, optional): figure. Defaults to None.
        figsize (list, optional): figure size. Defaults to None.
        ax_data (plt.Axes, optional): subplot with the data plot. Defaults to None.
        kws_seaborn (dict, optional): keyword parameters to seaborn plot. Defaults to {}.
        kws_annot_chroms (dict, optional): keyword parameters to the chromosome plot. Defaults to {}.
        kws_annot_labels (dict, optional): keyword parameters to the annotations of the labels. Defaults to {}.
        test (bool, optional): test mode. Defaults to False.

    Returns:
        plt.Figure: figure
    """
    if fig is None and not figsize is None:
        fig=plt.figure(figsize=figsize)
    else:
        fig=plt.gcf()
    if ax_data is None:
        ax_data=plt.subplot(polar=arc)
        ax_data_external=False
    else:
        ax_data_external=True
    if col_start is None and xkind=='loci':
        col_start=colx
        
    if not show_genome:
        cytobands=cytobands.query(expr=f"`chromosome` == {data['chromosome'].drop_duplicates().tolist()}")
    
    # if off is None:
    #     if not arc:
    #         off=0.9
    #     else:
    #         off=0.1
    data=data.astype({'chromosome':str})
    ## background
    ax_chrom,df1=annot_chroms(
        # ax=ax_data, # data
        data=cytobands,
        chromosomes=data['chromosome'].unique(),
        ax_chrom=None,
        chrom_y=chrom_y,
        out_data=True,
        pi_span=pi_span,
        # **kws_plot,
        kws_add_ax=dict(
            arc=arc,
            off=off,
            offy=offy,
            va=va,
            ax_with=ax_data,
            ),
        **kws_annot_chroms,
    )
    ax_chrom.set(ylim=[-0.6,0.1])
    # return
        
    if xkind=='loci':
        ## get coordinates of the labels on the joined chromosomes
        colx_data=colx+' chroms'
        col_start=colx_data
        df2=_concat_chroms(
                data,
                col_start=colx,
                col_end=colx,
                col_chroms_start=colx_data,
                col_chroms_end=colx_data,
                genome_ends=df1.set_index('chromosome')['genome end'],
            )
    else:
        colx_data=colx
        _col_start=col_start
        col_start+=' chroms'
        data=_concat_chroms(
                data,
                col_start=_col_start,
                col_end=_col_start,
                col_chroms_start=col_start,
                col_chroms_end=col_start,
                genome_ends=df1.set_index('chromosome')['genome end'],
            )
        
    if not ax_data_external:
        ## plot: middle
        ax_data,df3=plot_seaborn(
            df2,
            colx=colx_data,
            coly=coly,
            kind=kind,
            range1_chroms=[1, ## for polar
                           (cytobands
                           .groupby('chromosome',sort=False)['end']
                           .max().sum())
                          ],
            arc=arc,
            pi_span=pi_span,
            pi_start=pi_start,
            pi_end=pi_end,
            figsize=None,
            ax=ax_data,
            # fig=fig,
            **kws_seaborn,
            )
        df3=df3 if coffy is None else df3.query(expr=f"`{coly}` > {coffy}")
    else:
        if xkind=='loci':
            df3=df2.copy()
        else:
            df3=data.copy()            
    ax_data.set(zorder=1)
    ## labels: top
    ## filter for labels
    annot_labels(
        ax_chrom=ax_chrom, # A
        ax=ax_data, # B:data
        colx=colx_data,
        col_start=col_start,
        coly=coly,
        data=df3,
        chrom_y=chrom_y,
        col_label=col_label,
        fig=fig,
        **kws_annot_labels,
        test=test,
        )
    # return fig
    return {"chrom":ax_chrom,'data': ax_data,}

from functools import partial
plot_with_genome=partial(plot_with_chroms, show_genome=True,)


import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

from chrov.viz.chrom import _get_pi_range,to_polar,_format_polar_subplot,_concat_chroms

def plot_seaborn(
    data: pd.DataFrame,
    kind: str, # seaborn function name
    colx: str,
    coly: str,
    range1_chroms: list,
    arc: bool=True,
    pi_span: float=1,
    pi_start: int=0,
    pi_end: int=None,
    figsize: list=None,
    ax: plt.Axes=None,
    fig: plt.Figure=None,
    **kws_plot,
    ) -> tuple:
    """plot_seaborn _summary_

    Args:
        data (pd.DataFrame): input data
        kind (str): kind of plot, seaborn function name
        coly (str): column with y values
        range1_chroms (list): input range of chromosomes
        arc (bool, optional): arc/polar or linear/rectangular plots. Defaults to True.
        pi_span (float, optional): pi span. Defaults to 1.
        pi_start (int, optional): pi start position. Defaults to 0.
        pi_end (int, optional): pi end position. Defaults to None.
        figsize (list, optional): figure size. Defaults to None.
        ax (plt.Axes, optional): subplot. Defaults to None.
        fig (plt.Figure, optional): figure. Defaults to None.

    Returns:
        tuple: subplot and data

    TODOs:
        1. set rlabel position.
    """
    ## plot
    if ax is None:
        # if not arc:
        #     fig,ax=plt.subplots(figsize=figsize)
        #     df0=data.copy()
        # else:
        #     fig=plt.figure(figsize=figsize)
        if fig is None:
            fig=plt.gcf()
        ax=fig.add_axes(
            [0, 1, 1, 0.6],
            polar=arc,
            )
        ## TODO: in python>=3.8 with matplotlib>=3.7 used mosaic to make subplots: https://matplotlib.org/stable/gallery/subplots_axes_and_figures/mosaic.html#per-axes-subplot-keyword-arguments
    if ax.name=='polar':       
        xlim=_get_pi_range(pi_start=pi_start,pi_end=pi_end,pi_span=pi_span)
        
    df0=data.copy()
    if ax.name=='polar':
        # to_polar([df0[colx].min(),df0[colx].max()],
        #          range1=range1,
        #          range2=xlim,
        #         )
        df0=df0.assign(**{
            colx:lambda df: to_polar(df[colx],
                                     range1=range1_chroms,
                                     range2=xlim)
            # col_labelx:lambda df: to_polar(df[colx],range1=[df[colx].min(),df[colx].max()],range2=xlim)
        })
    if kind=='stem':
        ax.stem(df0[colx].tolist(),
                df0[coly].tolist(),
            basefmt='',
            **kws_plot,
            )
        if ax.name!='polar':
            ax.set(
                ylabel=coly,
                ylim=[kws_plot['bottom'] if 'bottom' in kws_plot else df0[coly].min(),df0[coly].max()]
            )        
    else:
        getattr(sns,kind)(
            data=df0,
            x=colx,
            y=coly,
            color='k',
            # clip_on = False,
            ax=ax,
            **kws_plot,
        )

    ax.set_clip_on(False)
    if ax.name!='polar':
        ax.margins(x=0.01)
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
    else:
        if pi_start==0 and pi_span==2:
            ax.set_rlabel_position(0)            
        if pi_start==0 and pi_span>=1:
            ax.set_rlabel_position(180)
        else:
            ax.set_rlabel_position(0)
            
        label_position=ax.get_rlabel_position()
        import numpy as np
        ax.text(
                x=np.radians(label_position),
                y=ax.get_rmax()/2.,
                s=coly,
                rotation=label_position if label_position!=180 else 0,
                ha='center',va='top'
               )
        _format_polar_subplot(ax=ax,
                              xlim=xlim,
                              # xlim=[xlim[0]-0.1,xlim[1]+0.1],
                             )
        ax.set(xticklabels=[],
               xlabel=None,ylabel=None,
              )
        ax.xaxis.grid(False)
        ax.yaxis.grid(color='whitesmoke',zorder=-1) # < lightgray connectors
        ax.spines['polar'].set_visible(False)
        ax.set_frame_on(False)
        # ax.text(0.25, 0.15, coly, transform=ax.transAxes,ha='center')
        
    return ax,df0

def heatmaps_strips(
    data: pd.DataFrame,
    strips_kws: dict,
    fig: plt.Figure=None,
    axs: list=None,
    kws_subplots: list={},
    ) -> tuple:
    """Plot heatmap strips

    Args:
        data (pd.DataFrame): input data
        strips_kws (dict): keyword arguments provided to strips
        fig (plt.Figure, optional): figure. Defaults to None.
        axs (list, optional): subplots. Defaults to None.
        kws_subplots (list, optional): keyword arguments provided to subplots. Defaults to {}.

    Returns:
        tuple: figure and subplots
    """
    if fig is None:
        fig,axs=plt.subplots(
            len(strips_kws),1,
            **kws_subplots
            )
    from roux.lib.str import linebreaker
    for axi,(ax,c) in enumerate(zip(axs,strips_kws.keys())):
        sns.heatmap(
            data.set_index('gene symbol label').loc[:,[c]].T,
            ax=ax,
            cbar_kws=dict(
                aspect=10,
            ),
            **strips_kws[c],
            )
        cbar = ax.collections[0].colorbar
        cbar.ax.tick_params(labelsize=11)
        cbar.set_label(
                linebreaker(c,20),
                       rotation=0,
                        # loc='center',
                        ha='left',
                        va='center',
                      )
        # cbar.ax.yaxis.set_label_coords(15,0.5)
        # cbar_ax.yaxis.label.set_size(20)
        ax.set(
            xlabel=None,
            ylabel=None,
            yticks=[],
            yticklabels=[],
        # **set_kws,
        )
        if len(data)>20 or axi!=len(strips_kws)-1:
            ax.set(        
                xticklabels=[]
            )        
    return fig,axs


"""Chromosome plots"""
import logging
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

import roux.lib.df as rd
from roux.stat.transform import rescale
from roux.viz.ax_ import get_line_cap_length

def _to_polar_intrapolate(
    start: int,
    end: int,
    interval: int=None,
    intervals: int=None,
    ) -> list:
    """Interpolate polar for smoothing.

    Args:
        start (int): start position
        end (int): end position
        interval (int, optional): interval size. Defaults to None.
        intervals (int, optional): number of intervals. Defaults to None.

    Raises:
        ValueError: either interval or intervals should be provided.

    Returns:
        list: interpolated positions
    """
    if not intervals is None:
        interval=(end-start)/intervals
    if interval is None and intervals is None:
        raise ValueError('either interval or intervals should be provided.')
    if end-start > interval:
        a1=np.append(np.arange(start,end,interval),end)
    else:
        a1=np.array([start,end])
    return a1

def _to_polar_chrom(
    range1_local: list,
    range1_global: list,    
    kws_plot_arm: dict,
    ax: plt.Axes,
    ):
    """To polar chromosome.

    Args:
        range1_local (list): local coordinates e.g. chrom
        range2_global (list): global coordinates e.g. genome
        kws_plot_arm (dict): keywords provided to `plot_arm` function
        ax (plt.Axes): plt.Axes
    """
    if ax.name=='polar':
        return {k:{k1:range1_local if k1=='range1' else rescale(a=range1_local,range1=range1_global,range2=kws_plot_arm['kws_pre_xys']['range2']) if k1=='range2' else v1 for k1,v1 in v.items()} if k=='kws_pre_xys' else v for k,v in kws_plot_arm.items()}
    else:
        return kws_plot_arm
    
def to_polar(
    a: list,
    range1: list=None,
    range2: list=None,
    interval: int=None,
    # test=False,
    ) -> list:
    """To polar coordinates

    Args:
        a (list): Coordinates
        range1 (list, optional): range1. Defaults to None.
        range2 (list, optional): range2. Defaults to None.
        interval (int, optional): interval size. Defaults to None.

    Raises:
        ValueError: Coordinates format

    Returns:
        list: rescaled coordinates
    """
    if not interval is None:
        a=sorted(a)
        if len(a)==2:
            start,end=a
        elif len(a)>2:
            start,end=a[0],a[-1]
        else:
            raise ValueError(a)
        a=_to_polar_intrapolate(
            start,
            end,
            interval,
        )
    range1=range1 if not range1 is None else [min(a),max(a)]
    return rescale(
        a,
        range1=range1,
        range2=[0,2*np.pi] if range2 is None else range2,
        )
            
def _pre_xys(
    xs: list,
    ys: list=None,
    polar: bool=False,
    **kws_to_polar,
    ) -> list:
    """Pre-process coordinates.

    Args:
        xs (list): x coordinates
        ys (list, optional): y coordinates. Defaults to None.
        polar (bool, optional): polar coordinates or not. Defaults to False.

    Returns:
        list: coordinates
    """
    if not polar:
        return xs,ys
    else:
        xs=to_polar(
            a=xs,
            **kws_to_polar,
        )
        assert len(set(ys))==1
        ys=np.repeat(ys[0],len(xs))
        return xs,ys
    
def _get_pi_range(
    pi_span: float=1,
    pi_start: int=0,
    pi_end: int=None,
    ) -> list:
    """Get pi range

    Args:
        pi_span (float, optional): pi span. Defaults to 1.
        pi_start (int, optional): pi start. Defaults to 0.
        pi_end (int, optional): pi end. Defaults to None.

    Returns:
        list: pi range
    """
    return [pi_start*np.pi,
            (pi_start+pi_span)*np.pi if pi_end is None else pi_end*np.pi]

def _format_polar_subplot(
    ax: plt.Axes,
    xlim: list,
    **kws_set
    ) -> plt.Axes:
    """Format polar subplot

    Args:
        ax (plt.Axes): plt.Axes
        xlim (list): x axis limits

    Returns:
        plt.Axes: subplot
    """
    if ax.name=='polar':
        if ax.get_theta_offset()!=np.pi:
            # ax.set_theta_zero_location("W")
            ax.set_theta_offset(np.pi)
        if ax.get_theta_direction()!=-1:
            ax.set_theta_direction(-1)      
        ax.set(xlim=xlim,**kws_set)
    return ax

def plot_arm(
    data: pd.DataFrame,
    arc: bool=False,
    col_start: str='start',
    col_end: str='end',
    y: float = 0,
    lw: float = 10,
    ec: str='k',
    color_centromer='#dbc1c1',
    pi_span: float=1,
    pi_start: int=0,
    pi_end: int=None,
    polar_smoothness_scale: float=1,
    kws_pre_xys: dict=None,
    figsize: list=None,
    ax: plt.Axes=None,
    test: bool=False,
    solid_capstyle='round',
    )->plt.Axes:
    """Plot chromosome arm.

    Args:
        data (pd.DataFrame): input table.
        y (float, optional): y position. Defaults to 0.
        lw (float, optional): line width. Defaults to 20.
        ec (str, optional): edge color. Defaults to 'k'.
        ax (plt.Axes, optional): subplot. Defaults to None.
        test (bool, optional): test-mode. Defaults to False.
    
    Returns:
        plt.Axes: subplot
    """
    ## subplot
    ax=_get_ax(ax=ax,arc=arc,figsize=figsize)
    
    if isinstance(data,dict):
        data=pd.DataFrame(data)
    assert data['chromosome'].nunique()==1, f"multiple chromosomes provided: {data['chromosome'].unique()}"
    start,end=data[col_start].min(),data[col_end].max()
    if ax is None:
        fig,ax=plt.subplots(figsize=[5,1])
    # to_set the line width
    if kws_pre_xys is None:
        kws_pre_xys={}
    ax.set(
        xlim=_pre_xys([start,end],['tmp'],**{k:v for k,v in kws_pre_xys.items() if k!='interval'})[0]
        )
    if solid_capstyle=='round':
        offx=get_line_cap_length(ax,linewidth=lw)[0]
    else:
        offx=0
    # print('kws_pre_xys',kws_pre_xys)
    if ax.name=='polar':
        kws_pre_xys['polar']=True
        if solid_capstyle!='butt':
            logging.info("For curcular/polar plot, `solid_capstyle` can only be set to 'butt'. Other styles are currently not supported.")
            solid_capstyle='butt'
        if not 'range1' in kws_pre_xys:
            kws_pre_xys['range1']=start,end
        if not 'range2' in kws_pre_xys:
            assert pi_span>0 and pi_span<=2 
            kws_pre_xys['range2']=_get_pi_range(pi_span=pi_span,pi_start=pi_start,pi_end=pi_end)
        if not 'interval' in kws_pre_xys:
            kws_pre_xys['interval']=(end-start)/(100*polar_smoothness_scale)
    else:
        kws_pre_xys['polar']=False
    if not 'cytoband type' in data:
        logging.warning('cytoband type not found in the data')
    else:
        ## cytobands
        data.loc[data['cytoband type'].str.startswith('gpos'),:].apply(lambda x: ax.plot(
            *_pre_xys(
                [x[col_start],x[col_end]],[y,y],**kws_pre_xys),

            lw=lw,solid_capstyle='butt',
            alpha=int(x['cytoband type'].replace('gpos',''))*0.01,
            color='k',zorder=2,clip_on=False,
            ),axis=1)
        data.loc[data['cytoband type'].str.startswith('gneg'),:].apply(lambda x: ax.plot(
            *_pre_xys(
                [x[col_start],x[col_end]],[y,y],**kws_pre_xys),

            lw=lw,solid_capstyle='butt',
            color='w',alpha=1,
            zorder=1,clip_on=False,
            ),axis=1)
        ## centromere
        data.loc[(data['cytoband type']=='acen'),:].log().apply(lambda x: ax.plot(
            *_pre_xys([x[col_start],x[col_end]],[y,y],**kws_pre_xys),

            lw=lw*0.75,solid_capstyle='butt',
            color=color_centromer,#'lightcoral', ##F5B8B7
            zorder=2,clip_on=False,
            ),axis=1)
        ## outlines
        start_line,end_line=data.query("`cytoband type`!='acen'").agg({col_start:'min',col_end:'max'}).tolist()
        ax.plot(
            *_pre_xys([start_line+offx,end_line-offx],[y,y],**kws_pre_xys),
            lw=lw+2, solid_capstyle=solid_capstyle,
            color=ec,zorder=1,clip_on=False
            )
        ax.plot(
            *_pre_xys([start_line+offx,end_line-offx],[y,y],**kws_pre_xys),
            lw=lw, solid_capstyle=solid_capstyle,
            color='w',zorder=1,clip_on=False
            )
    if not test:
        ax.axis('off')
    _format_polar_subplot(
        ax=ax,
        xlim=kws_pre_xys['range2'] if 'range2' in kws_pre_xys else None,
    )
    return ax

def plot_chrom(
    data: pd.DataFrame,
    arc: bool=False,
    col_start: str='start',
    col_end: str='end',
    col_arm: str='arm',
    pi_span: float=1,
    pi_start: int=0,
    pi_end: int=None,
    ax: plt.Axes=None,
    figsize: list=None,
    **kws_plot_arm,
    ) -> plt.Axes:
    """Plot a chromosome

    Args:
        data (pd.DataFrame): cytobands
        arc (bool, optional): arc/polar mode. Defaults to False.
        col_start (str, optional): column with start positions. Defaults to 'start'.
        col_end (str, optional): column with end positions. Defaults to 'end'.
        col_arm (str, optional): column with arm names. Defaults to 'arm'.
        pi_span (float, optional): pi span. Defaults to 1.
        pi_start (int, optional): pi start. Defaults to 0.
        pi_end (int, optional): pi end. Defaults to None.
        ax (plt.Axes, optional): subplot. Defaults to None.
        figsize (list, optional): sigure size. Defaults to None.

    Returns:
        plt.Axes: subplot
    """
    ## subplot
    ax=_get_ax(ax=ax,arc=arc,figsize=figsize)
    if not 'kws_pre_xys' in kws_plot_arm:
        kws_plot_arm['kws_pre_xys']={}
    if not 'range2' in kws_plot_arm['kws_pre_xys']:
        assert pi_span>0 and pi_span<=2 
        kws_plot_arm['kws_pre_xys']['range2']=_get_pi_range(pi_span=pi_span,pi_start=pi_start,pi_end=pi_end)
    _=data.sort_values(col_start).groupby(col_arm,sort=False,observed=False).apply(lambda df: plot_arm(
        data=df,
        col_start=col_start,
        col_end=col_end,
        **_to_polar_chrom(
            range1_local=[df[col_start].min(),df[col_end].max()],
            range1_global=[data[col_start].min(),data[col_end].max()],
            kws_plot_arm=kws_plot_arm,
            ax=ax,
        ),
        ax=ax
        ))
    if ax.name=='polar':
        try:
            ax.set(xlim=rescale(
                [data[col_start].min(),data[col_end].max()],
                range2=kws_plot_arm['kws_pre_xys']['range2'],
            ))    
        except:
            print(
            [data[col_start].min(),data[col_end].max()],
             kws_plot_arm['kws_pre_xys']['range2'],
        )

    return ax

def _rescale_to_chroms(
    df: pd.DataFrame,
    col_start: str,
    col_chrom_start: str,
    col_chroms_start: str,
    col_end: str=None,
    col_chroms_end: str=None,
    ) -> pd.DataFrame:
    """Rescale coordinates to joined chromosomes.

    Args:
        df (pd.DataFrame): cytobands
        col_start (str): column with gene start position
        col_chrom_start (str): column with chromosome start position
        col_chroms_start (str): column with chromosomes start position
        col_end (str, optional): column with end position. Defaults to None.
        col_chroms_end (str, optional): column with chromosomes end position. Defaults to None.

    Returns:
        pd.DataFrame: rescaled positions
    """
    df=df.assign(
        **{
            col_chroms_start:lambda df: df.apply(lambda x: x[col_chrom_start]-1+x[col_start]  ,axis=1),
        }
        )
    if not col_chroms_end is None:
        df=df.assign(
            **{
                col_chroms_end:lambda df: df.apply(lambda x: x[col_chrom_start]-1+x[col_end]  ,axis=1),
            }
            )
    return df

def _sort_chroms(
    data: pd.DataFrame,
    col_start: str=None,
    chromosomes: list=None,
    ) -> pd.DataFrame:
    """Sort chromosome.

    Args:
        data (pd.DataFrame): cytoband
        col_start (str, optional): column with start position. Defaults to None.
        chromosomes (list, optional): list of chromosomes. Defaults to None.

    Returns:
        pd.DataFrame: table with sorted chromosomes.
    """
    def _format_chrom_names(x): return int(x) if isinstance(x,float) else x if isinstance(x,int) else int(x) if x.isnumeric() else x
    ## sort chroms
    data=(data
        .assign(
            **{
            'chromosome': lambda df: df['chromosome'].apply(
                # lambda x: int(x) if x.isnumeric() else x
                _format_chrom_names,
            ),
            'chrom name instance': lambda df: df['chromosome'].apply(lambda x: not isinstance(x,int)),
        }
        )
        .sort_values(['chrom name instance','chromosome'],ascending=[True,True])
    )
    chromosomes_all=data['chromosome'].drop_duplicates().tolist()
    if chromosomes is None:
        # sort chromosomes
        chromosomes=list(map(_format_chrom_names,chromosomes_all))
        if len(chromosomes)>1 and len(chromosomes)!=len(chromosomes_all):
            logging.warning(f"chromosomes order not provided. inferred one: {chromosomes}")
    else:
        chromosomes=list(map(lambda x: _format_chrom_names(x), chromosomes))
        from roux.lib.set import assert_overlaps_with
        assert_overlaps_with(chromosomes,chromosomes_all)
        ## sort
        chromosomes=[k for k in chromosomes_all if k in chromosomes]
        ## filter
        data=(data.log.query(expr=f"`chromosome` == {chromosomes}"))   
    assert len(data)!=0, (chromosomes)
    # print(data)
    data.rd.assert_no_na(subset=['chromosome'])
    
    # sort loci
    data=(data
        .rd.astype_cat(col='chromosome', cats=chromosomes)
        .sort_values(['chromosome']+([] if col_start is None else [col_start]))
        )
    return data

def _concat_chroms(
    data: pd.DataFrame,
    col_start: str,
    col_end: str,
    col_chroms_start: str,
    col_chroms_end: str,
    col_chrom_start: str='col_chrom_start',
    col_chrom_end: str='col_chrom_end',
    # chromosomes=None,
    genome_ends: pd.Series=None, ## reference genome ends based on cytobands for e.g. 
    test: bool=False,
    ) -> pd.DataFrame:
    """Concat chromosomes by calculating global start and end positions.

    Args:
        data (pd.DataFrame): cytobands for multiple chromosomes
        col_start (str): column with start positions
        col_end (str): column with end positions
        col_chroms_start (str): column with start positions for chromosomes
        col_chroms_end (str): column with end positions for chromosomes
        col_chrom_start (str, optional): column with start positions for a chromosome. Defaults to 'col_chrom_start'.
        col_chrom_end (str, optional): column with end positions for a chromosome. Defaults to 'col_chrom_end'.
        genome_ends (pd.Series, optional): genome end positions. Defaults to None.

    Returns:
        pd.DataFrame: concatenated positions of the chromosomes
    """
    data=_sort_chroms(
        data,
        col_start,
        chromosomes=None,
    )
    ## end positions in the order along genome (contatenated chroms)
    if genome_ends is None:
        genome_ends=data.groupby('chromosome',sort=False,observed=False)[col_end].max().cumsum()
        
    ### .. of previous chromosome
    _genome_ends=genome_ends.shift(fill_value=0).to_dict()

    if test:
        print(data.groupby('chromosome',sort=False,observed=False)[col_end].max())
        print(_genome_ends)

    df1=(data
        .assign(
            **{
                'chrom i': lambda df: df.groupby('chromosome',sort=False,observed=False).ngroup(), 
                col_chrom_start: lambda df: df['chromosome'].map(_genome_ends).astype(int)+1,
                col_chrom_end: lambda df: df['chromosome'].map(genome_ends),        
            }
        )
        .pipe(lambda df: _rescale_to_chroms(
             df,
             col_start=col_start,col_end=col_end,
             col_chrom_start=col_chrom_start,
             col_chroms_start=col_chroms_start,col_chroms_end=col_chroms_end,
       ))
       )
    return df1

def _get_ax(
    ax: plt.Axes,
    arc: bool,
    **kws_subplots,
    ) -> plt.Axes:
    """Get subplot object

    Args:
        ax (plt.Axes): subplot
        arc (bool): arc/polar mode.

    Returns:
        plt.Axes: subplot
    """
    if ax is None:
        _,ax=plt.subplots(
        subplot_kw=dict(projection='polar' if arc else  'rectilinear'), 
        **kws_subplots,
        )
    return ax

def _set_span_polar(
    ax: plt.Axes,
    span_type: str,
    start: int,
    end: int,
    intervals: int=5,
    **kws,
    ):
    """Set span polar

    Args:
        ax (plt.Axes): subplot
        span_type (str): span type
        start (int): start position
        end (int): end position
        intervals (int, optional): number of intervals. Defaults to 5.
    """
    intervals=_to_polar_intrapolate(
        start=start,
        end=end,
        intervals=intervals, ## parameterise
        )
    for xrange in np.column_stack([intervals[:-1], intervals[1:]]):
        getattr(ax,f"ax{span_type}span")(
            *xrange,
            **kws
            )

def plot_chroms(
    data: pd.DataFrame,
    arc=True,
    chromosomes: list=None,
    col_start='start',
    col_end='end',
    col_arm='arm',
    span_color: str ='#dcdcdc',
    span_color_alpha_scale: float =1,
    pi_span: float=1,
    pi_start: int=0,
    pi_end: int=None,
    show_labels: bool=True,
    show_vline: bool=True,
    label_y: str=None,
    test: bool=False,
    ax: plt.Axes=None,
    figsize: list=None,
    out_data: bool=False,
    **kws_plot_arm,    
    ):
    """Plot chromosomes joined.

    Args:
        data (pd.DataFrame): cytonbands
        arc (bool, optional): arc/polar mode. Defaults to True.
        chromosomes (list, optional): chromosomes. Defaults to None.
        col_start (str, optional): column with start position. Defaults to 'start'.
        col_end (str, optional): column with end position. Defaults to 'end'.
        col_arm (str, optional): column with chromosome arm names. Defaults to 'arm'.
        span_color (str, optional): span color. Defaults to 'whitesmoke'.
        span_color_alpha_scale (float, optional): span color transparency scale. Defaults to 1.
        pi_span (float, optional): pi span. Defaults to 1.
        pi_start (int, optional): pi start angle. Defaults to 0.
        pi_end (int, optional): pi end angle. Defaults to None.
        show_labels (bool, optional): show labels. Defaults to True.
        show_vline (bool, optional): show vertical line. Defaults to True.
        label_y (str, optional): label y. Defaults to None.
        test (bool, optional): test-mode. Defaults to False.
        ax (plt.Axes, optional): subplot. Defaults to None.
        figsize (list, optional): figure size. Defaults to None.
        out_data (bool, optional): output data. Defaults to False.
    """
    ## subplot
    ax=_get_ax(ax=ax,arc=arc,figsize=figsize)
    
    if not 'kws_pre_xys' in kws_plot_arm:
        kws_plot_arm['kws_pre_xys']={}
    if not 'range2' in kws_plot_arm['kws_pre_xys']:
        if ax.name=='polar':
            assert pi_span>0 and pi_span<=2, pi_span
            kws_plot_arm['kws_pre_xys']['range2']=_get_pi_range(pi_span=pi_span,pi_start=pi_start,pi_end=pi_end)
    df1=_concat_chroms(
        data,
        col_start=col_start,
        col_end=col_end,
        col_chrom_start='genome chrom start',
        col_chrom_end='genome chrom end',
        col_chroms_start='genome start',
        col_chroms_end='genome end',
        test=test,
        )
    
    start,end=df1.agg({'genome start':'min','genome end':'max'}).tolist()
    ## chromosomes
    _=(df1
       .groupby('chromosome',sort=False,observed=False)
       .apply(lambda df: plot_chrom(
            data=df,
            col_start='genome start',
            col_end='genome end',
            col_arm=col_arm,
            solid_capstyle='butt' if df1['chromosome'].nunique()>1 else 'round',
            **_to_polar_chrom(
                range1_local=[df['genome start'].min(),df['genome end'].max()],
                range1_global=[start,end],
                kws_plot_arm=kws_plot_arm.copy(),
                ax=ax,
            ),
            test=test,
            ax=ax
        ))
      )

    ## vspans
    _xlim=df1.agg({'genome start':'min','genome end':'max'}).tolist()
    ax.set(xlim=_xlim if not ax.name=='polar' else rescale(_xlim,range2=kws_plot_arm['kws_pre_xys']['range2']))

    df2=df1.groupby('chromosome',sort=False,observed=False).agg({'genome start':'min','genome end':'max','chrom i':'min'}).reset_index()
    if ax.name=='polar':
        df_=df2.apply(lambda x: rescale([x['genome start'],x['genome end']],range1=[start,end],range2=kws_plot_arm['kws_pre_xys']['range2'],),axis=1).apply(pd.Series)
        df_.columns=['genome start theta','genome end theta']
        df2=df2.join(df_)
        if show_vline and (kws_plot_arm['kws_pre_xys']['range2'][1]-kws_plot_arm['kws_pre_xys']['range2'][0])%(2*np.pi) != 0:
            ## vline
            ax.axvline(x=kws_plot_arm['kws_pre_xys']['range2'][0],color='k')
            ax.axvline(x=kws_plot_arm['kws_pre_xys']['range2'][1],color='k')
    if ax.name!='polar':
        _=(df2
            .apply(lambda x: ax.axvspan(
                *[x["genome start"],x["genome end"]],
                color=span_color,
                alpha=span_color_alpha_scale*(x['chrom i']%2),
                zorder=-1,
                ),
            axis=1)
        )    
    else:
        _=(df2
            .apply(lambda x: _set_span_polar(
                ax=ax,
                span_type='v',
                start=x["genome start theta"],
                end=x["genome end theta"],
                color=span_color,
                alpha=span_color_alpha_scale*(x['chrom i']%2),
                zorder=-1,
            ),
            axis=1)
        )    
    
    if show_labels:
        if label_y is None:
            ## infer label_y 
            if 'y' in kws_plot_arm:
                label_y=kws_plot_arm['y']
            else:
                if ax.name=='polar':
                    label_y=0.1*1.5
                else:
                    label_y=0.1*0.8
        df2.apply(lambda x: ax.text(
            x=np.mean([x[f"genome start{' theta' if ax.name=='polar' else ''}"],x[f"genome end{' theta' if ax.name=='polar' else ''}"]]),
            y=label_y,
            s=x['chromosome'] if isinstance(x['chromosome'],(int,str)) else f"{x['chromosome']:.0f}",
            ha='center',
            va='center' if ax.name=='polar' else 'bottom',
        ),axis=1)
    if out_data:
        return ax,df2
    else:
        return ax
    
def _add_ax(
    off: float,
    va: str,
    offy: float=None,
    arc: bool=None,
    fig: plt.Figure=None,
    rect: list=None,
    ax_with: plt.Axes=None,
    zorder: int=0,
    # test: bool=False,
    ) -> plt.Axes:
    """Create the subplot for the chromosomes.


    Args:
        off (float): offset scale
        va (str): vertical alignment
        offy (float, optional): offset y. Defaults to None.
        arc (bool, optional): arc/polar mode. Defaults to None.
        fig (plt.Figure, optional): figure. Defaults to None.
        rect (list, optional): rectangle positions. Defaults to None.
        ax_with (plt.Axes, optional): subplot to relate with. Defaults to None.
        zorder (int, optional): z-order. Defaults to 0.

    Returns:
        plt.Axes: subplot
    """
    assert not (ax_with is None and arc is None)
    if fig is None:
        fig=plt.gcf()    
    if not ax_with is None:
        rect=ax_with.get_position().bounds
        if arc is None:
            arc=ax_with.name=='polar'
    if offy is None:
        offy=off
    if not rect is None:
        if va=='center':
            rect=[ 
                rect[0]+(off*-0.5),
                rect[1]+(offy*-0.5),
                rect[2]+off,
                rect[3]+off,
                ]
        elif va=='center bottom':
            rect=[
                rect[0],
                rect[1]+offy,
                rect[2]+off,
                rect[3]+off,
                ]
        elif va=='bottom':
            rect=[
                rect[0],
                rect[1]+offy,
                rect[2],
                rect[3]+off,
                ]
        else:
            raise ValueError(va)
    else:
        ## fail-safe inference
        if arc:
            ## experimental
            if offy==0:
                rect=[
                    -(off*0.5),
                    -(off*0.5),
                    1+off,
                    1+off,
                ] #(left, bottom, width, height)
            else:
                if offy is None:
                    offy=off*0.5
                rect=[-1*off,
                      (1-offy),
                      1+(2*off),
                      1+(2*off)]
        else:
            rect=[0,1,1,1*(off+0.3)]
    # if test:
    logging.info(f"rect={rect}")
    ax=fig.add_axes(
        rect=rect,#(left, bottom, width, height),
        zorder=zorder,
        polar=arc,
        )
    return ax

def annot_chroms(
    data: pd.DataFrame,
    chromosomes: list,
    ax_chrom: plt.Axes=None,
    chrom_y: float=0,
    kws_add_ax: dict={},
    test: bool=False,
    **kws_plot,
    ) -> plt.Axes:
    """Add a subplot with the chromosome.


    Args:
        data (pd.DataFrame): table with cytobands
        chromosomes (list): chromosomes
        ax_chrom (plt.Axes, optional): subplot with chromosome plot. Defaults to None.
        chrom_y (float, optional): chromosome y-position. Defaults to 0.
        kws_add_ax (dict, optional): keyword parameters provided to `_add_ax`. Defaults to {}.
        test (bool, optional): test mode. Defaults to False.

    Returns:
        plt.Axes: subplot
    """
    ## chromosomes
    if ax_chrom is None:            
        ax_chrom=_add_ax(
            # fig,
            # arc=arc,
            # scale=scale,
            # rect=rect,
            # offy=offy,
            **kws_add_ax,
            )
    
    ax=plot_chroms(
        data=data,
        chromosomes=chromosomes,
        show_labels=True,
        test=test,
        y=chrom_y,
        label_y=.05,
        ax=ax_chrom,
        # out_data=True,
        show_vline=False,
        **kws_plot,
        )
    if not isinstance(ax,tuple):
        _=ax.set(ylim=[-1,0.1])
    else:
        _=ax[0].set(ylim=[-1,0.1])
    # ax.ticklabel_format(style='plain')
    return ax